import copy
import pdb, math

from components.episode_buffer import EpisodeBatch
from modules.critics.ader import FACMACCritic, ADERCritic_Conti
import torch as th
from torch.optim import RMSprop, Adam
from modules.mixers.vdn import VDNMixer
from modules.mixers.qmix import QMixer
from modules.mixers.ader import ADERMixer, ADERMixer_Ent
import numpy as np


from torch.autograd import Variable

# from modules.mixers.qmix_ablations import VDNState, QMixerNonmonotonic

alpha_min = 1e-4
alpha_max =  1


class ADER_conti_learner:
    def __init__(self, mac, scheme, logger, args):
        self.args = args
        self.n_agents = args.n_agents
        self.n_actions = args.n_actions
        self.logger = logger

        self.mac = mac
        self.target_mac = copy.deepcopy(self.mac)
        self.agent_params = list(mac.parameters())



        self.critic = ADERCritic_Conti(scheme, args)
        self.critic2 = ADERCritic_Conti(scheme, args)
        self.critic_ent = ADERCritic_Conti(scheme, args)
        self.critic_ent2 = ADERCritic_Conti(scheme, args)

        self.target_critic = copy.deepcopy(self.critic)
        self.target_critic2 = copy.deepcopy(self.critic2)
        self.target_critic_ent = copy.deepcopy(self.critic_ent)
        self.target_critic_ent2 = copy.deepcopy(self.critic_ent2)
        self.critic_params = list(self.critic.parameters()) + list(self.critic_ent.parameters())
        self.critic_params2= list(self.critic2.parameters()) + list(self.critic_ent2.parameters())



        self.mixer = None
        if args.mixer is not None and self.args.n_agents > 1:  # if just 1 agent do not mix anything
            if args.mixer == "vdn":
                self.mixer = VDNMixer()
            elif args.mixer == "qmix":
                self.mixer = ADERMixer(args)
                self.mixer2 = ADERMixer(args)
                self.mixer_ent = ADERMixer_Ent(args)
                self.mixer_ent2 = ADERMixer_Ent(args)
            elif args.mixer == "vdn-s":
                self.mixer = VDNState(args)
            elif args.mixer == "qmix-nonmonotonic":
                self.mixer = QMixerNonmonotonic(args)
            else:
                raise ValueError("Mixer {} not recognised.".format(args.mixer))
            self.target_mixer = copy.deepcopy(self.mixer)
            self.target_mixer2 = copy.deepcopy(self.mixer2)
            self.target_mixer_ent = copy.deepcopy(self.mixer_ent)
            self.target_mixer_ent2 = copy.deepcopy(self.mixer_ent2)

            self.critic_params += list(self.mixer.parameters()) + list(self.mixer_ent.parameters())
            self.critic_params2 += list(self.mixer2.parameters()) + list(self.mixer_ent2.parameters())




        if getattr(self.args, "optimizer", "rmsprop") == "rmsprop":
            self.agent_optimiser = RMSprop(params=self.agent_params, lr=args.lr, alpha=args.optim_alpha,
                                           eps=args.optim_eps)
        elif getattr(self.args, "optimizer", "rmsprop") == "adam":
            self.agent_optimiser = Adam(params=self.agent_params, lr=args.lr,
                                        eps=getattr(args, "optimizer_epsilon", 10E-8))
        else:
            raise Exception("unknown optimizer {}".format(getattr(self.args, "optimizer", "rmsprop")))

        if getattr(self.args, "optimizer", "rmsprop") == "rmsprop":
            self.critic_optimiser = RMSprop(params=self.critic_params, lr=args.critic_lr, alpha=args.optim_alpha,
                                            eps=args.optim_eps)
            self.critic_optimiser2 = RMSprop(params=self.critic_params2, lr=args.critic_lr, alpha=args.optim_alpha,
                                            eps=args.optim_eps)
        elif getattr(self.args, "optimizer", "rmsprop") == "adam":
            self.critic_optimiser = Adam(params=self.critic_params, lr=args.critic_lr,
                                         eps=getattr(args, "optimizer_epsilon", 10E-8))
            self.critic_optimiser2 = Adam(params=self.critic_params2, lr=args.critic_lr,
                                         eps=getattr(args, "optimizer_epsilon", 10E-8))
        else:
            raise Exception("unknown optimizer {}".format(getattr(self.args, "optimizer", "rmsprop")))

        self.log_stats_t = -self.args.learner_log_interval - 1

        self.log_alpha = Variable(self.args.adap_total_alpha_start * th.ones(self.n_agents).cuda(),
                                  requires_grad=True)  # .cuda()
        self.alpha = th.exp(self.log_alpha)
        self.target_entropy_sum = np.ones(self.n_agents) * (-self.n_actions)

        self.alpha_params = [self.log_alpha]
        if self.args.optimizer == "Adam":
            self.alpha_optimiser = Adam(params=self.alpha_params, lr=args.critic_lr)
        else:
            self.alpha_optimiser = RMSprop(params=self.alpha_params, lr=args.critic_lr, alpha=args.optim_alpha,
                                           eps=args.optim_eps)

    def train(self, batch: EpisodeBatch, t_env: int, episode_num: int):
        # Get the relevant quantities
        rewards = batch["reward"][:, :-1]
        actions = batch["actions"][:, :-1]
        terminated = batch["terminated"][:, :-1].float()
        mask = batch["filled"][:, :-1].float()
        mask[:, 1:] = mask[:, 1:] * (1 - terminated[:, :-1])
        self.alpha = th.exp(self.log_alpha)
        alpha = self.alpha.detach()
        alpha = th.clamp(alpha, max=alpha_max, min=alpha_min).detach()
        self.alpha = alpha

        # Train the actor
        # Optimize over the entire joint action space
        mac_out, entropy = [], []
        chosen_action_qvals, chosen_action_qvals_ent = [], []
        q_total_each_grad, q_total_sigma_grad = [], []
        self.mac.init_hidden(batch.batch_size)
        self.critic.init_hidden(batch.batch_size)
        self.critic2.init_hidden(batch.batch_size)
        for t in range(batch.max_seq_length):
            forward_info = self.mac.forward(batch, t=t, select_actions=True)
            agent_outs = forward_info["actions"].view(batch.batch_size, self.n_agents, self.n_actions)
            q = self.critic(self._build_inputs(batch, t=t), agent_outs, self.critic.hidden_states).view(batch.batch_size, -1, 1)
            q2 = self.critic2(self._build_inputs(batch, t=t), agent_outs, self.critic2.hidden_states).view(
                batch.batch_size, -1, 1)
            q_ent = self.critic_ent(self._build_inputs(batch, t=t), agent_outs, self.critic_ent.hidden_states)
            q_ent2 = self.critic_ent2(self._build_inputs(batch, t=t), agent_outs, self.critic_ent2.hidden_states)
            if self.mixer is not None:
                q_m, q_grad = self.mixer(q, batch["state"][:, t:t + 1], grad=True)
                q_m2, q_grad2 = self.mixer2(q2, batch["state"][:, t:t + 1], grad=True)
                q_ent = self.mixer_ent(q_ent.view(batch.batch_size, -1, 1), batch["state"][:, t:t + 1])
                q_ent2 = self.mixer_ent2(q_ent2.view(batch.batch_size, -1, 1), batch["state"][:, t:t + 1])

            q_tot_min_ind = th.unsqueeze(th.argmin(th.cat([q_m, q_m2], dim=2), dim=-1), dim=-1).repeat(1, 1, self.args.n_agents)
            contribution = q_tot_min_ind * q_grad + (1 - q_tot_min_ind) * q_grad2

            grad11 = th.autograd.grad(q_m.sum(), agent_outs, retain_graph=True)[0].squeeze(dim=-1).view(
                batch.batch_size, self.n_actions, self.args.n_agents)
            grad21 = th.autograd.grad(forward_info["actions"].sum(), forward_info["log_sigma"], retain_graph=True)[0]
            grad21 = grad21.view(batch.batch_size, self.n_actions, self.args.n_agents)
            grad31 = (grad11 * grad21)

            grad12 = th.autograd.grad(q_m2.sum(), agent_outs, retain_graph=True)[0].squeeze(dim=-1).view(
                batch.batch_size, self.n_actions, self.args.n_agents)
            grad22 = th.autograd.grad(forward_info["actions"].sum(), forward_info["log_sigma"], retain_graph=True)[0]
            grad22 = grad22.view(batch.batch_size, self.n_actions, self.args.n_agents)
            grad32 = (grad12 * grad22)

            grad3 = q_tot_min_ind * grad31 + (1 - q_tot_min_ind) * grad32

            q_total_sigma_grad.append(grad3[:, 0, :])



            mac_out.append(agent_outs)
            chosen_action_qvals.append(th.min(q_m, q_m2))
            # q_total_each_grad.append(q_grad[:, 0, :])
            q_total_each_grad.append(contribution[:, 0, :])

            chosen_action_qvals_ent.append(th.min(q_ent,q_ent2))
            entropy.append(-1 * forward_info["logp_pi"].view(batch.batch_size, -1))

        mac_out = th.stack(mac_out[:-1], dim=1)
        q_total_sigma_grad = th.stack(q_total_sigma_grad[:-1], dim=1)
        chosen_action_qvals = th.stack(chosen_action_qvals[:-1], dim=1)
        q_total_each_grad = th.stack(q_total_each_grad[:-1], dim=1)

        entropy = th.stack(entropy[:-1], dim=1)#.sum(dim=-1)
        chosen_action_qvals_ent = th.stack(chosen_action_qvals_ent[:-1], dim=1)

        w_entropy = entropy * alpha
        w_chosen_action_qvals_ent = (chosen_action_qvals_ent * alpha).sum(dim=-1)

        # Compute the actor loss
        pg_loss = -(chosen_action_qvals.mean() + w_entropy.mean() + w_chosen_action_qvals_ent.mean())

        # Optimise agents
        self.agent_optimiser.zero_grad()
        pg_loss.backward()
        agent_grad_norm = th.nn.utils.clip_grad_norm_(self.agent_params, self.args.grad_norm_clip)
        self.agent_optimiser.step()

        # Train the critic batched
        target_actions = []
        self.target_mac.init_hidden(batch.batch_size)
        for t in range(batch.max_seq_length):
            agent_target_outs = self.target_mac.select_actions(batch, t_ep=t, t_env=None, test_mode=True,
                                                               critic=self.target_critic, target_mac=True)
            target_actions.append(agent_target_outs)
        target_actions = th.stack(target_actions, dim=1)  # Concat over time

        q_taken, q_taken_ent, q_taken2, q_taken_ent2 = [], [], [], []
        # self.critic.init_hidden(batch.batch_size)
        for t in range(batch.max_seq_length - 1):
            inputs = self._build_inputs(batch, t=t)
            critic_out  = self.critic(inputs, actions[:, t:t + 1].detach())
            critic_out_ent = self.critic_ent(inputs, actions[:, t:t + 1].detach())
            critic_out2 = self.critic2(inputs, actions[:, t:t + 1].detach())
            critic_out_ent2 = self.critic_ent2(inputs, actions[:, t:t + 1].detach())
            if self.mixer is not None:
                critic_out_m = self.mixer(critic_out.view(batch.batch_size, -1, 1), batch["state"][:, t:t + 1])
                critic_out_ent = self.mixer_ent(critic_out_ent.view(batch.batch_size, -1, 1), batch["state"][:, t:t + 1])
                critic_out_m2 = self.mixer2(critic_out2.view(batch.batch_size, -1, 1), batch["state"][:, t:t + 1])
                critic_out_ent2 = self.mixer_ent2(critic_out_ent2.view(batch.batch_size, -1, 1),
                                                batch["state"][:, t:t + 1])
            q_taken.append(critic_out_m)
            q_taken_ent.append(critic_out_ent)
            q_taken2.append(critic_out_m2)
            q_taken_ent2.append(critic_out_ent2)
        q_taken = th.stack(q_taken, dim=1)
        q_taken_ent = th.stack(q_taken_ent, dim=1)
        q_taken2 = th.stack(q_taken2, dim=1)
        q_taken_ent2 = th.stack(q_taken_ent2, dim=1)


        target_vals, target_vals_ent, target_vals_ent2, target_vals2 =[], [], [], []
        self.target_critic.init_hidden(batch.batch_size)
        self.target_critic2.init_hidden(batch.batch_size)
        for t in range(1, batch.max_seq_length):
            target_inputs = self._build_inputs(batch, t=t)
            target_critic_out = self.target_critic(target_inputs, target_actions[:, t:t + 1].detach(),
                                                                  self.target_critic.hidden_states)
            target_critic_out_ent = self.target_critic_ent(target_inputs, target_actions[:, t:t + 1].detach(),
                                                   self.target_critic_ent.hidden_states)
            target_critic_out2 = self.target_critic2(target_inputs, target_actions[:, t:t + 1].detach(),
                                                   self.target_critic.hidden_states)
            target_critic_out_ent2 = self.target_critic_ent2(target_inputs, target_actions[:, t:t + 1].detach(),
                                                           self.target_critic_ent2.hidden_states)
            if self.mixer is not None:
                target_critic_out = self.target_mixer(target_critic_out.view(batch.batch_size, -1, 1),
                                                      batch["state"][:, t:t + 1])
                target_critic_out_ent = self.target_mixer_ent(target_critic_out_ent.view(batch.batch_size, -1, 1),batch["state"][:, t:t + 1])
                target_critic_out2 = self.target_mixer2(target_critic_out2.view(batch.batch_size, -1, 1),
                                                      batch["state"][:, t:t + 1])
                target_critic_out_ent2 = self.target_mixer_ent2(target_critic_out_ent2.view(batch.batch_size, -1, 1),
                                                              batch["state"][:, t:t + 1])
            target_vals.append(target_critic_out)
            target_vals_ent.append(target_critic_out_ent)
            target_vals2.append(target_critic_out2)
            target_vals_ent2.append(target_critic_out_ent2)
        target_vals = th.stack(target_vals, dim=1)
        target_vals_ent = th.stack(target_vals_ent, dim=1)
        target_vals2 = th.stack(target_vals2, dim=1)
        target_vals_ent2 = th.stack(target_vals_ent2, dim=1)

        if self.mixer is not None:
            q_taken = q_taken.view(batch.batch_size, -1, 1)
            q_taken_ent = q_taken_ent.view(batch.batch_size, -1, self.n_agents)
            q_taken2 = q_taken2.view(batch.batch_size, -1, 1)
            q_taken_ent2 = q_taken_ent2.view(batch.batch_size, -1, self.n_agents)
            target_vals = target_vals.view(batch.batch_size, -1, 1)
            target_vals_ent = target_vals_ent.view(batch.batch_size, -1, self.n_agents)
            target_vals2 = target_vals2.view(batch.batch_size, -1, 1)
            target_vals_ent2 = target_vals_ent2.view(batch.batch_size, -1, self.n_agents)
        else:
            q_taken = q_taken.view(batch.batch_size, -1, self.n_agents)
            target_vals = target_vals.view(batch.batch_size, -1, self.n_agents)

        target_vals= th.min(target_vals, target_vals2)
        targets = rewards.expand_as(target_vals) + self.args.gamma * (1 - terminated.expand_as(target_vals)) * target_vals

        rew_ent = th.zeros_like(entropy)
        rew_ent[:, 1:] = entropy[:, :-1].detach() * self.args.gamma

        target_ent_vals = th.min(target_vals_ent, target_vals_ent2)
        targets_ent = rew_ent + self.args.gamma * (1 - terminated.expand_as(target_ent_vals)) * target_ent_vals

        td_error = (targets.detach() - q_taken)
        td_error_ent = (targets_ent.detach() - q_taken_ent)
        td_error2 = (targets.detach() - q_taken2)
        td_error_ent2 = (targets_ent.detach() - q_taken_ent2)

        mask = mask.expand_as(td_error)
        masked_td_error = td_error * mask
        masked_td_error_ent = td_error_ent * mask
        masked_td_error2 = td_error2 * mask
        masked_td_error_ent2 = td_error_ent2 * mask

        loss = (masked_td_error ** 2).sum() / mask.sum() + (masked_td_error_ent ** 2).sum() / (self.n_agents * mask.sum())


        self.critic_optimiser.zero_grad()
        loss.backward()
        critic_grad_norm = th.nn.utils.clip_grad_norm_(self.critic_params, self.args.grad_norm_clip)
        self.critic_optimiser.step()

        loss2 = (masked_td_error2 ** 2).sum() / mask.sum() + (masked_td_error_ent2 ** 2).sum() / (
                    self.n_agents * mask.sum())
        self.critic_optimiser2.zero_grad()
        loss2.backward()
        critic_grad_norm = th.nn.utils.clip_grad_norm_(self.critic_params2, self.args.grad_norm_clip)
        self.critic_optimiser2.step()

        ### alpha ###

        exploration_num = 1
        weight_v_ent = th.zeros_like(td_error_ent)
        q_total_grad = -q_total_sigma_grad.detach()
        if self.args.contribution_temperature == -100:
            q_total_each_grad2 = q_total_grad / (
                    th.unsqueeze(th.max(abs(q_total_grad), dim=-1)[0], dim=-1) + 1e-5)
            contribution_softmax = th.softmax(q_total_each_grad2 / 1, dim=2)
        elif self.args.contribution_temperature == -1000:
            q_total_each_grad2 = q_total_grad / (
                    th.unsqueeze(th.max(q_total_grad, dim=-1)[0], dim=-1) * 0.1 + 1e-5)
            contribution_softmax = th.softmax(q_total_each_grad2 / 1, dim=2)
        else:
            q_total_each_grad2 = q_total_grad / (
                    th.unsqueeze(th.max(abs(q_total_grad), dim=-1)[0],
                                 dim=-1) * self.args.contribution_temperature2 + 1e-5)
            contribution_softmax = th.softmax(q_total_each_grad2 / 1, dim=2)
        weight_v_ent = weight_v_ent + contribution_softmax

        weight_v_ent_ = (weight_v_ent / exploration_num).detach()
        weight_v_ent_softmax_avg = th.mean(weight_v_ent_, dim=1).mean(dim=0)
        self.target_entropy_sum = (self.args.adap_total_alpha_tau) * self.target_entropy_sum + (
                1 - self.args.adap_total_alpha_tau) * weight_v_ent_softmax_avg.cpu().detach().numpy() * sum(
            self.target_entropy_sum)

        entropy = entropy.reshape(-1, self.args.n_agents)
        target_entropy = th.ones_like(entropy)
        for i in range(self.n_agents):
            target_entropy[:, i] = target_entropy[:, i] * self.target_entropy_sum[i]
        alpha_backup = target_entropy - entropy
        alpha_loss = -(self.log_alpha * alpha_backup.detach()).sum(dim=1).mean()

        self.alpha_optimiser.zero_grad()
        alpha_loss.backward()
        agent_grad_norm = th.nn.utils.clip_grad_norm_(self.alpha_params, self.args.grad_norm_clip)
        self.alpha_optimiser.step()



        if getattr(self.args, "target_update_mode", "hard") == "hard":
            self._update_targets()
        elif getattr(self.args, "target_update_mode", "hard") in ["soft", "exponential_moving_average"]:
            self._update_targets_soft(tau=getattr(self.args, "target_update_tau", 0.001))
        else:
            raise Exception(
                "unknown target update mode: {}!".format(getattr(self.args, "target_update_mode", "hard")))

        if t_env - self.log_stats_t >= self.args.learner_log_interval:
            self.logger.log_stat("critic_loss", loss.item(), t_env)
            self.logger.log_stat("critic_grad_norm", critic_grad_norm.item(), t_env)
            mask_elems = mask.sum().item()
            self.logger.log_stat("target_mean", targets.sum().item() / mask_elems, t_env)
            self.logger.log_stat("pg_loss", pg_loss.item(), t_env)
            self.logger.log_stat("agent_grad_norm", agent_grad_norm.item(), t_env)
            self.log_stats_t = t_env
            for i in range(self.n_agents):
                self.logger.log_stat("target_entropy"+ str(i), self.target_entropy_sum[i].item(), t_env)
                self.logger.log_stat("alpha" + str(i), self.alpha[i].item(), t_env)

            # entropy

    def _update_targets_soft(self, tau):
        for target_param, param in zip(self.target_mac.parameters(), self.mac.parameters()):
            target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)

        for target_param, param in zip(self.target_critic.parameters(), self.critic.parameters()):
            target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)

        for target_param, param in zip(self.target_critic_ent.parameters(), self.critic_ent.parameters()):
            target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)

        for target_param, param in zip(self.target_critic2.parameters(), self.critic2.parameters()):
            target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)

        for target_param, param in zip(self.target_critic_ent2.parameters(), self.critic_ent2.parameters()):
            target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)

        if self.mixer is not None:
            for target_param, param in zip(self.target_mixer.parameters(), self.mixer.parameters()):
                target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)
            for target_param, param in zip(self.target_mixer_ent.parameters(), self.mixer_ent.parameters()):
                target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)
            for target_param, param in zip(self.target_mixer2.parameters(), self.mixer2.parameters()):
                target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)
            for target_param, param in zip(self.target_mixer_ent2.parameters(), self.mixer_ent2.parameters()):
                target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)

        if self.args.verbose:
            self.logger.console_logger.info("Updated all target networks (soft update tau={})".format(tau))

    def _build_inputs(self, batch, t):
        bs = batch.batch_size
        inputs = []
        inputs.append(batch["obs"][:, t])
        if self.args.obs_agent_id:
            inputs.append(th.eye(self.n_agents, device=batch.device).unsqueeze(0).expand(bs, -1, -1))


        inputs = th.cat([x.reshape(bs * self.n_agents, -1) for x in inputs], dim=1)
        return inputs

    def _update_targets(self):
        self.target_mac.load_state(self.mac)
        self.target_critic.load_state_dict(self.critic.state_dict())
        self.target_critic_ent.load_state_dict(self.critic_ent.state_dict())
        self.target_critic2.load_state_dict(self.critic2.state_dict())
        self.target_critic_ent2.load_state_dict(self.critic_ent2.state_dict())
        if self.mixer is not None:
            self.target_mixer.load_state_dict(self.mixer.state_dict())
            self.target_mixer_ent.load_state_dict(self.mixer_ent.state_dict())
            self.target_mixer2.load_state_dict(self.mixer2.state_dict())
            self.target_mixer_ent2.load_state_dict(self.mixer_ent2.state_dict())
        self.logger.console_logger.info("Updated all target networks")

    def cuda(self, device="cuda:0"):
        self.mac.cuda()
        self.target_mac.cuda()
        self.critic.cuda()
        self.target_critic.cuda()
        self.critic_ent.cuda()
        self.target_critic_ent.cuda()

        self.critic2.cuda()
        self.target_critic2.cuda()
        self.critic_ent2.cuda()
        self.target_critic_ent2.cuda()

        if self.mixer is not None:
            self.mixer.cuda()
            self.target_mixer.cuda()
            self.mixer_ent.cuda()
            self.target_mixer_ent.cuda()

            self.mixer2.cuda()
            self.target_mixer2.cuda()
            self.mixer_ent2.cuda()
            self.target_mixer_ent2.cuda()
        self.alpha.cuda()

    def save_models(self, path):
        self.mac.save_models(path)
        if self.mixer is not None:
            th.save(self.mixer.state_dict(), "{}/mixer.th".format(path))
        th.save(self.agent_optimiser.state_dict(), "{}/opt.th".format(path))

    def load_models(self, path):
        self.mac.load_models(path)
        # Not quite right but I don't want to save target networks
        self.target_mac.load_models(path)
        if self.mixer is not None:
            self.mixer.load_state_dict(th.load("{}/mixer.th".format(path), map_location=lambda storage, loc: storage))
        self.agent_optimiser.load_state_dict(
            th.load("{}/opt.th".format(path), map_location=lambda storage, loc: storage))